from cmath import pi
import json
import jsonpickle
from json import JSONEncoder
import pickle
import numpy as np
import numba
from numba import jit
import scipy.io as sio
import h5py

Timepoint = np.dtype([('TR', np.float32), ('TE', np.float32), ('FA', np.float32)])
Beat = np.dtype([('RRTime', np.float32), ('PrepTime', np.float32), ('PrepID', np.float32)])

class Sequence:

    def __init__(self, name, timepoints=[], beats=[]):
        self.name = name
        self.timepoints = timepoints
        self.beats = beats

    def __str__(self):
        return "(" + str(self.timepoints) + ", " + str(self.beats) + ")"

    def Initialize(self, TRs, TEs, FAs, rrTimes, prepTimes, prepIDs):
        self.timepoints = np.empty(len(TRs), dtype=Timepoint)
        if (len(TRs)!=len(TEs)) or (len(TRs)!=len(FAs)):
            print("Import Failed: TR/TE/FA files must have identical number of entries")
            return 
        for index in range(len(TRs)):
            self.timepoints[index] = (TRs[index], TEs[index], FAs[index])
        self.beats = np.empty(len(rrTimes), dtype=Beat)
        if (len(rrTimes)!=len(prepTimes)) or (len(rrTimes)!=len(prepIDs)):
            print("Import Failed: RRTime/PrepTime/PrepID files must have identical number of entries")
            return 
        for index in range(len(rrTimes)):
            self.beats[index] = (rrTimes[index], prepTimes[index], prepIDs[index])
        print(self.name + " || Initialized " + str(len(self.timepoints)) + " timepoint definitions and " + str(len(self.beats)) + " beat definitions")

    def ImportFromTxt(self, trFilepath, teFilepath, faFilepath, rrTimeFilepath, prepTimeFilepath, prepIDFilepath):
        TRs = np.loadtxt(trFilepath)
        TEs = np.loadtxt(teFilepath)
        FAs = np.loadtxt(faFilepath)
        rrTimes = np.loadtxt(rrTimeFilepath)
        prepTimes = np.loadtxt(prepTimeFilepath)
        prepIDs = np.loadtxt(prepIDFilepath)
        self.Initialize(TRs, TEs, FAs, rrTimes, prepTimes, prepIDs)

    def Export(self, filename, force=False):
        if ".mrf" in filename:
            outfile = h5py.File(filename, "a")
            try:
                sequencesGroup = outfile.create_group("sequences")
            except:
                print("Sequences group already exists. Appending.") 
            if (self.name in list(outfile["sequences"].keys())) and not force:
                print("Sequence already exists in .mrf file. Specify 'force' to overwrite")
            else:
                sequence = sequencesGroup.create_group(self.name)
                timepointDataset = sequence.create_dataset("timepoints",shape=len(self.timepoints), dtype=Timepoint)
                timepointDataset.write_direct(self.timepoints)
                beatDataset = sequence.create_dataset("beats",shape=len(self.beats), dtype=Beat)
                beatDataset.write_direct(self.beats)           
                outfile.close()
        else:
            print("Input is not a .mrf file")

    @staticmethod
    def Import(filename, sequenceName):
        if ".mrf" in filename:
            infile = h5py.File(filename, "r")
            sequenceGroup = infile["sequences"][sequenceName]
            new_sequence = Sequence(sequenceName, sequenceGroup["timepoints"][:], sequenceGroup["beats"][:])
            infile.close()
            return new_sequence
        else:
            print("Input is not a .mrf file")
    
    @staticmethod
    def GetAvailableSequences(filename):
        if ".mrf" in filename:
            infile = h5py.File(filename, "r")
            return list(infile["sequences"].keys())
        else:
            print("Input is not a .mrf file")


from cmath import pi
import jsonpickle
import numpy as np
import numba
from numba import jit
import scipy.io as sio
import h5py

DictionaryEntry = np.dtype([('T1', np.float32), ('T2', np.float32), ('B1', np.float32)])

class Dictionary:
    def __init__(self, name, entries=[]):
        self.name = name
        self.entries = entries

    def __str__(self):
        return "(" + str(len(self.entries)) + ")"

    def Initialize(self, T1s, T2s, B1s=[]):
        self.entries = np.empty(len(T1s), dtype=DictionaryEntry)
        if (len(T1s)!=len(T2s)):
            print("Import Failed: T1/T2 files must have identical number of entries")
            return 
        for index in range(len(T1s)):
            if(B1s != []):
                for b1Index in range(len(B1s)):
                    self.entries[index] = (T1s[index], T2s[index], B1s[b1Index])
            else:
                self.entries[index] = (T1s[index], T2s[index], None)
        print(self.name + " || Populated " + str(len(self.entries)) + " entries")

    def ImportFromTxt(self, T1Filepath, T2Filepath, B1Filepath=""):
        T1s = np.loadtxt(T1Filepath)
        T2s = np.loadtxt(T2Filepath)
        if(B1Filepath != ""):
            B1s = np.loadtxt(B1Filepath)
            self.Initialize(T1s, T2s, B1s)
        else:
            self.Initialize(T1s,T2s)
            
    def PopulateFromPercent(self, t1Range=(100,4000), t2Range=(1,400), percentStepSize=5, includeB1=False, b1Range=(0.5,1.5), b1Stepsize=0.05):
        T1s = []
        T2s = []
        t1 = t1Range[0]
        t2 = t2Range[0]
        while t1 <= t1Range[1]:
            T1s.append(t1/1000)
            t1 = t1*(1+(percentStepSize/100))
        while t2 <= t2Range[1]:
            T2s.append(t2/1000)
            t2 = t2*(1+(percentStepSize/100))
        pairs = []
        for t1Val in T1s:
            for t2Val in T2s:
                if(t1Val>t2Val): # Don't include pairs with T2 longer than T1
                    pairs.append((t1Val,t2Val))
        T1sFromPairs = []
        T2sFromPairs = []
        for pair in pairs:
            T1sFromPairs.append(pair[0])
            T2sFromPairs.append(pair[1])
        if(includeB1):
            B1s = np.arange(b1Range[0], b1Range[1], b1Stepsize)
            self.Initialize(T1sFromPairs,T2sFromPairs, B1s)
        else:
            self.Initialize(T1sFromPairs, T2sFromPairs)

    def PopulateFromFixedStep(self, t1Range=(100,4000), t2Range=(1,400), fixedStepSize=1, includeB1=False, b1Range=(0.5,1.5), b1Stepsize=0.05):
        T1s = []
        T2s = []
        t1 = t1Range[0]
        t2 = t2Range[0]
        while t1 <= t1Range[1]:
            T1s.append(t1/1000)
            t1 = t1+fixedStepSize
        while t2 <= t2Range[1]:
            T2s.append(t2/1000)
            t2 = t2+fixedStepSize
        pairs = []
        for t1Val in T1s:
            for t2Val in T2s:
                if(t1Val>t2Val): # Don't include pairs with T2 longer than T1
                    pairs.append((t1Val,t2Val))
        T1sFromPairs = []
        T2sFromPairs = []
        for pair in pairs:
            T1sFromPairs.append(pair[0])
            T2sFromPairs.append(pair[1])
        if(includeB1):
            B1s = np.arange(b1Range[0], b1Range[1], b1Stepsize)
            self.Initialize(T1sFromPairs,T2sFromPairs, B1s)
        else:
            self.Initialize(T1sFromPairs, T2sFromPairs)

    def SaveToTxtFiles(self, includeB1=False):
        t1File = open(self.name+"_T1s.txt","w")
        t2File = open(self.name+"_T2s.txt","w")
        for entry in self.entries:
            t1File.write(f'{entry["T1"]:7.5f}'+"\n")
            t2File.write(f'{entry["T2"]:7.5f}'+"\n") 
        if(includeB1):
            b1File = open(self.name+"_B1s.txt","w")
            for entry in self.entries:
                b1File.write(f'{entry["B1"]:7.5f}'+"\n") 
    
    def Export(self, filename, force=False):
        if ".mrf" in filename:
            outfile = h5py.File(filename, "a")
            try:
                dictionariesGroup = outfile.create_group("dictionaries")
            except:
                print("Dictionary group already exists. Appending.") 
            if (self.name in list(outfile["dictionaries"].keys())) and not force:
                print("Dictionary already exists in .mrf file. Specify 'force' to overwrite")
            else:
                dictionary = dictionariesGroup.create_group(self.name)
                timepointDataset = dictionary.create_dataset("entries",shape=len(self.entries), dtype=DictionaryEntry)
                timepointDataset.write_direct(self.entries)
                outfile.close()
        else:
            print("Input is not a .mrf file")

    @staticmethod
    def Import(filename, dictionaryName):
        if ".mrf" in filename:
            infile = h5py.File(filename, "r")
            dictionaryGroup = infile["dictionaries"][dictionaryName]
            new_dictionary = Dictionary(dictionaryName, dictionaryGroup["entries"][:])
            infile.close()
            return new_dictionary
        else:
            print("Input is not a .mrf file")
    
    @staticmethod
    def GetAvailableDictionaries(filename):
        if ".mrf" in filename:
            infile = h5py.File(filename, "r")
            return list(infile["dictionaries"].keys())
        else:
            print("Input is not a .mrf file")


from cmath import pi
import jsonpickle
import numpy as np
import numba
from numba import jit
import scipy.io as sio

SimulationConfiguration = np.dtype([('sequence', np.unicode_, 32), ('dictionary', np.unicode_, 32), ('phaseRange', np.float32,(2,)), ('spins', np.int64), ('averageSpins', np.bool8)])

class Simulation: 
    def __init__(self, sequence, dictionary, simulationConfiguration):
        self.sequence = sequence
        self.dictionary = dictionary
        self.simulationConfiguration = simulationConfiguration

    @staticmethod
    def GenerateConfiguration(sequenceName, dictionaryName, phaseRange=(-8*np.pi, 8*np.pi), spins=1, averageSpins=False):
        config = np.empty(1,dtype=SimulationConfiguration)[0]
        config['sequence'] = sequenceName
        config['dictionary'] = dictionaryName
        config['phaseRange'] = phaseRange
        config['spins'] = spins
        config['averageSpins'] = averageSpins
        return config

    @staticmethod
    @jit(parallel=True, nopython=True)
    def ExecuteNumbaSimulation(numBeats,numTimepointsPerBeat,T1s,T2s,TRs,TEs,FAs,prepIDs,prepTimes,rrTimes,phaseValues,numberOfSpins):

        numberOfExperiments = np.int32(numTimepointsPerBeat*numBeats)         
        numberOfDictionaryEntries = len(T1s)
        FAs = FAs[0:numberOfExperiments]
        prepTimes = prepTimes/1000
        TREs = TRs-TEs        
        Mx0 = np.zeros((numberOfExperiments,numberOfDictionaryEntries,numberOfSpins))
        My0 = np.zeros((numberOfExperiments,numberOfDictionaryEntries,numberOfSpins))
        Mz0 = np.zeros((numberOfExperiments,numberOfDictionaryEntries,numberOfSpins))

        phaseValueCosines = np.cos(phaseValues)
        phaseValueSines = np.sin(phaseValues)
                            
        for dictionaryEntryNumber in numba.prange(0,numberOfDictionaryEntries):
            
            T1 = T1s[dictionaryEntryNumber]
            T2 = T2s[dictionaryEntryNumber]
        
            Mx = np.zeros(numberOfSpins)
            My = np.zeros(numberOfSpins)
            Mz = np.zeros(numberOfSpins)
            Mz[:] = 1
            
            for ibeat in range(numBeats):
                baseExperimentNumber = ibeat*numTimepointsPerBeat
                tr = TRs[baseExperimentNumber]
                te = TEs[baseExperimentNumber]
                tre = TREs[baseExperimentNumber]

                At2te = np.exp(-1*te/T2)
                At1te = np.exp(-1*te/T1)
                Bt1te = 1-At1te
                
                At2tr = np.exp(-1*tre/T2)
                At1tr = np.exp(-1*tre/T1)
                Bt1tr = 1-At1tr

                RRtime = rrTimes[ibeat]
                At2rr = np.exp(-1*RRtime/T2)
                At1rr = np.exp(-1*RRtime/T1)
                Bt1rr = 1-At1rr
                
                Preptime = prepTimes[ibeat]
                At2p = np.exp(-1*Preptime/T2)
                At1p = np.exp(-1*Preptime/T1)
                Bt1p = 1-At1p
                
                Mx = np.multiply(Mx,At2rr)
                My = np.multiply(My,At2rr)
                Mz = np.multiply(Mz,At1rr)+Bt1rr
                    
                if prepIDs[ibeat] == 1:
                    My = -My
                    Mz = -Mz
                    Mx = np.multiply(Mx,At2p)
                    My = np.multiply(My,At2p)
                    Mz = np.multiply(Mz,At1p)+Bt1p
                    Mxi = Mx
                    Myi = My
                    Mx = np.multiply(phaseValueCosines,Mxi) - np.multiply(phaseValueSines,Myi)
                    My = np.multiply(phaseValueSines,Mxi) + np.multiply(phaseValueCosines,Myi)
                    
                elif prepIDs[ibeat] == 2:
                    Myi = My
                    Mzi = Mz
                    My = -Mzi
                    Mz = Myi
                    Mx = np.multiply(Mx,At2p)
                    My = np.multiply(My,At2p)
                    Mz = np.multiply(Mz,At1p)+Bt1p
                    Myi = My
                    Mzi = Mz
                    My = Mzi
                    Mz = -Myi
                    Mxi = Mx
                    Myi = My
                    Mx = np.multiply(phaseValueCosines,Mxi) - np.multiply(phaseValueSines,Myi)
                    My = np.multiply(phaseValueSines,Mxi) + np.multiply(phaseValueCosines,Myi)
                    
                for timepointNumber in range(numTimepointsPerBeat):
                    experimentNumber = baseExperimentNumber+timepointNumber
                    crf = np.cos(FAs[experimentNumber])
                    srf = np.sin(FAs[experimentNumber])
                    Myi = My
                    Mzi = Mz
                    My = np.multiply(crf,Myi)-np.multiply(srf,Mzi)
                    Mz = np.multiply(srf,Myi)+np.multiply(crf,Mzi)
                    Mx = np.multiply(Mx,At2te)
                    My = np.multiply(My,At2te)
                    Mz = np.multiply(Mz,At1te)+Bt1te
                    Mx0[experimentNumber,dictionaryEntryNumber,:]=Mx[:]
                    My0[experimentNumber,dictionaryEntryNumber,:]=My[:]
                    Mz0[experimentNumber,dictionaryEntryNumber,:]=Mz[:]
                    Mx = Mx*At2tr
                    My = My*At2tr
                    Mz = Mz*At1tr+Bt1tr
                    Mxi = Mx
                    Myi = My
                    Mx = np.multiply(phaseValueCosines,Mxi) - np.multiply(phaseValueSines,Myi)
                    My = np.multiply(phaseValueSines,Mxi) + np.multiply(phaseValueCosines,Myi)
        return Mx0,My0,Mz0
        
    @staticmethod
    @jit(parallel=True, nopython=True)
    def ParallizedMeans(Mx0,My0,Mz0):
        MeansXo = np.zeros((np.shape(Mx0)[0],np.shape(Mx0)[1]))
        MeansYo = np.zeros((np.shape(Mx0)[0],np.shape(Mx0)[1]))
        MeansZo = np.zeros((np.shape(Mx0)[0],np.shape(Mx0)[1]))

        for n in numba.prange(np.shape(Mx0)[1]):
            for timepoint in range(np.shape(Mx0)[0]):
                MeansXo[timepoint,n] = np.mean(Mx0[timepoint,n,:])
                MeansYo[timepoint,n] = np.mean(My0[timepoint,n,:])
                MeansZo[timepoint,n] = np.mean(Mz0[timepoint,n,:])
        return MeansXo, MeansYo, MeansZo

    def Execute(self, numBeats=-1, numTimepointsPerBeat=-1):
        if(numBeats == -1):
            numBeats = len(self.sequence.beats)
        if(numTimepointsPerBeat == -1):
            numTimepointsPerBeat = int(len(self.sequence.timepoints)/numBeats)

        TRs = []
        TEs = []
        FAs = []
        prepIDs = []
        prepTimes = []
        rrTimes = []
        T1s= []
        T2s = []
        B1s = []

        for dictionaryEntry in self.dictionary.entries:
            T1s.append(dictionaryEntry['T1'])
            T2s.append(dictionaryEntry['T2'])
            B1s.append(dictionaryEntry['B1'])

        for timepoint in self.sequence.timepoints:
            TRs.append(timepoint['TR'])
            TEs.append(timepoint['TE'])
            FAs.append(timepoint['FA'])

        for beat in self.sequence.beats:
            prepIDs.append(beat['PrepID'])
            prepTimes.append(beat['PrepTime'])
            rrTimes.append(beat['RRTime'])
    
        phaseValues = np.linspace(-self.simulationConfiguration['phaseRange'][0]/2,self.simulationConfiguration['phaseRange'][1]/2,self.simulationConfiguration['spins'])

        Mx0,My0,Mz0 = self.ExecuteNumbaSimulation(numTimepointsPerBeat,numBeats,np.asarray(T1s),np.asarray(T2s),np.asarray(TRs),np.asarray(TEs),np.asarray(FAs),np.asarray(prepIDs),np.asarray(prepTimes),np.asarray(rrTimes),phaseValues,self.simulationConfiguration['spins'])
       
        if(self.simulationConfiguration['averageSpins']):
            simulationResults = self.ParallizedMeans(Mx0,My0,Mz0)
        else:
            simulationResults = (Mx0, My0, Mz0)
        
        return simulationResults


